package no.ntnu.item.arctis.examples.android.geirsag.rtp.rtptransceiver;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.DatagramSocket;
import java.net.InetSocketAddress;
import java.net.SocketException;

import jlibrtp.DataFrame;
import jlibrtp.DebugAppIntf;
import jlibrtp.Participant;
import jlibrtp.RTCPAppIntf;
import jlibrtp.RTPAppIntf;
import jlibrtp.RTPSession;
import no.ntnu.item.arctis.examples.android.geirsag.rtp.objects.DestinationAddress;
import no.ntnu.item.arctis.examples.android.geirsag.rtp.objects.RTPTransceiverBundle;
import no.ntnu.item.arctis.runtime.android.AndroidBlock;
import android.util.Log;

public class RTPTransceiver extends AndroidBlock {

	public static final int maxBufferSize = 1480;
	public static final boolean DEBUG_MODE = true;

	private int rtpPktCounter = 0;
	private int rtcpPktCounter = 0;

	private InputStream inputStream;
	private OutputStream outputStream;
	private DatagramSocket rtpSocket;
	private DatagramSocket rtcpSocket;
	private RTPSession rtpSession;
	private volatile boolean isSending = false;

	private Thread senderThread = new Thread() {
		public void run() {
			isSending = true;
			byte[] buffer = new byte[maxBufferSize];
			int bytesRead;
			while (isSending) {
				try {
					bytesRead = inputStream.read(buffer);
					if (bytesRead < 0)
						isSending = false;
					else if (bytesRead < maxBufferSize) {
						byte[] croppedBuffer = new byte[bytesRead];
						for (int i = 0; i < croppedBuffer.length; i++) {
							croppedBuffer[i] = buffer[i];
						}
						rtpSession.sendData(croppedBuffer);
					} else {
						rtpSession.sendData(buffer);
					}
				} catch (IOException e) {
					e.printStackTrace();
					isSending = false;
				}
			}
			try {
				inputStream.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	};

	public boolean initialize(RTPTransceiverBundle bundle) {
		int rtpPort = bundle.receiverPort;
		int rtcpPort = rtpPort + 1;
		try {
			rtpSocket = new DatagramSocket(rtpPort);
			rtcpSocket = new DatagramSocket(rtcpPort);
			Log.d("RTP", "Sockets open at ports: " + rtpPort + " and " + rtcpPort);
		} catch (SocketException e) {
			Log.d("RTP", "Could not open sockets at ports: " + rtpPort + " and " + rtcpPort);
			e.printStackTrace();
			return false;
		}

		this.inputStream = bundle.inputStream;
		this.outputStream = bundle.outputStream;

		rtpSession = new RTPSession(rtpSocket, rtcpSocket);
		rtpSession.naivePktReception(true);
		rtpSession.RTPSessionRegister(new RTPHandler(), new RTCPHandler(), DEBUG_MODE ? new DebugHandler() : null);

		senderThread.start();

		return true;

	}

	public void stopSession() {
		isSending = false;
		try {
			senderThread.join();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		rtpSession.endSession();
		rtpSocket.close();
		rtcpSocket.close();
	}

	private class RTPHandler implements RTPAppIntf {

		@Override
		public void receiveData(DataFrame frame, Participant participant) {
			try {
				outputStream.write(frame.getConcatenatedData());
			} catch (IOException e) {
				Log.e("IOException",
						"Error writing received RTP frame to output stream, message: " + e.getLocalizedMessage(), e);
			}
		}

		@Override
		public void userEvent(int type, Participant[] participant) {

		}

		@Override
		public int frameSize(int payloadType) {
			return 1;
		}
	}

	private class RTCPHandler implements RTCPAppIntf {

		@Override
		public void SRPktReceived(long ssrc, long ntpHighOrder, long ntpLowOrder, long rtpTimestamp, long packetCount,
				long octetCount, long[] reporteeSsrc, int[] lossFraction, int[] cumulPacketsLost, long[] extHighSeq,
				long[] interArrivalJitter, long[] lastSRTimeStamp, long[] delayLastSR) {
			// TODO Auto-generated method stub

		}

		@Override
		public void RRPktReceived(long reporterSsrc, long[] reporteeSsrc, int[] lossFraction, int[] cumulPacketsLost,
				long[] extHighSeq, long[] interArrivalJitter, long[] lastSRTimeStamp, long[] delayLastSR) {
			// TODO Auto-generated method stub

		}

		@Override
		public void SDESPktReceived(Participant[] relevantParticipants) {
			// TODO Auto-generated method stub

		}

		@Override
		public void BYEPktReceived(Participant[] relevantParticipants, String reason) {
			// TODO Auto-generated method stub
			StringBuilder builder = new StringBuilder("BYE packet received from:\n");
			for (Participant p : relevantParticipants) {
				builder.append("SSRC: " + p.getSSRC() + "  CNAME: " + p.getCNAME() + "\n");
			}
			Log.i("RTCP", builder.toString());
		}

		@Override
		public void APPPktReceived(Participant part, int subtype, byte[] name, byte[] data) {
			// TODO Auto-generated method stub

		}

	}

	private class DebugHandler implements DebugAppIntf {

		private static final int PACKET_INTERVAL = 100;

		@Override
		public void packetReceived(int type, InetSocketAddress socket, String description) {
			if (type == 0 && (++rtpPktCounter % PACKET_INTERVAL) == 0)
				Log.d("RTP", rtpPktCounter + " RTP packets received.");
			else if (type == 1)
				Log.d("RTCP", ++rtcpPktCounter + " RTCP packets received.");
			else
				Log.d("RTP/RTCP", "Invalid packet received: " + description);
		}

		@Override
		public void packetSent(int type, InetSocketAddress socket, String description) {
			if (type == 0 && (++rtpPktCounter % PACKET_INTERVAL) == 0)
				Log.d("RTP", rtpPktCounter + " RTP packets sent.");
			else if (type == 1)
				Log.d("RTCP", ++rtcpPktCounter + " RTCP packets sent.");
			else
				Log.d("RTP/RTCP", "Invalid packet received: " + description);
		}

		@Override
		public void importantEvent(int type, String description) {
			Log.d("RTP", "Important event: " + description);
		}

	}

	public void connect(DestinationAddress destination) {
		String address = destination.address;
		int port = destination.port;

		Participant p = new Participant(address, port, port + 1);
		rtpSession.addParticipant(p);
		isSending = true;
	}

}
